/* _NVRM_COPYRIGHT_BEGIN_
 *
 * Copyright 2001-2002 by NVIDIA Corporation.  All rights reserved.  All
 * information contained herein is proprietary and confidential to NVIDIA
 * Corporation.  Any use, reproduction, or disclosure without the written
 * permission of NVIDIA Corporation is prohibited.
 *
 * _NVRM_COPYRIGHT_END_
 */

#include "nv-misc.h"
#include "os-interface.h"
#include "nv.h"
#include "nv-freebsd.h"


/*
 * The NVIDIA kernel module's malloc identifier, needed for both tracking
 * and actual allocation/freeing purposes. M_NVIDIA is declared elsewhere
 * to make it known to other parts of the kernel module (nv-freebsd.h).
 */

MALLOC_DEFINE(M_NVIDIA, "nvidia", "NVIDIA memory allocations");

/*
 * The global client information array; it is used in the open and in the
 * closed source sections of the driver module. The core resource manager
 * was compiled for NV_MAX_CLIENTS also, modifying this define will yield
 * undefined results.
 */

nv_client_t nv_clients[NV_MAX_CLIENTS];


#define MAX_ERROR_STRING    256


RM_STATUS os_alloc_contig_pages(
    void **address,
    U032 size
)
{
    *address = contigmalloc(size, M_NVIDIA, 0, 0, ~0, PAGE_SIZE, 0);
    return *address ? RM_OK : RM_ERROR;
}

void os_free_contig_pages(
    void *address,
    U032 size
)
{
    contigfree(address, size, M_NVIDIA);
}

void* os_map_user_space(
    void *address,
    void **private,
    U032 size,
    U032 mode
)
{
    /*
     * XXX I don't think that there's a good way (any?) to implement this
     * function on FreeBSD. It is not currently used except by the NVIDIA
     * AGP mapping code, which we do not use on FreeBSD, but if it should
     * ever be called from elsewhere, we need to know.
     */
    os_dbg_breakpoint();
    return NULL;
}

void os_unmap_user_space(
    void *address,
    void *private
)
{
    /*
     * XXX I don't think that there's a good way (any?) to implement this
     * function on FreeBSD. It is not currently used except by the NVIDIA
     * AGP mapping code, which we do not use on FreeBSD, but if it should
     * ever be called from elsewhere, we need to know.
     */
    os_dbg_breakpoint();
}

/*
 * The core resource manager's favorite source of memory, this routine is
 * called from different contexts, including ISRs. This means that it can
 * not be allowed to sleep when memory is low.
 */

RM_STATUS os_alloc_mem(
    void **address,
    U032 size
)
{
    /* XXX Fix me? (malloc flags) */
    *address = malloc(size, M_NVIDIA, M_NOWAIT|M_ZERO);
    return *address ? RM_OK : RM_ERROR;
}

void os_free_mem(void *address)
{
    free(address, M_NVIDIA);
}


RM_STATUS os_delay(U032 MilliSeconds)
{
    DELAY(MilliSeconds * 1000);
    return RM_OK;
}

RM_STATUS os_delay_us(U032 MicroSeconds)
{
    DELAY(MicroSeconds);
    return RM_OK;
}

U032 os_get_cpu_frequency(void)
{
    /* XXX Fix me! (x86 only) */
    return ((tsc_freq + 4999) / 1000000);
}

RM_STATUS os_get_current_process(U032 *pPid)
{
    *pPid = curproc->p_pid;
    return RM_OK;
}

RM_STATUS os_kill_process(
    U032 pid,
    U032 sig
)
{
    struct proc *p = pfind(pid);
    if (!p) {
        /*
         * No process with the specified PID was found running. We notify
         * the caller, but take no other action.
         */
        return RM_ERR_OPERATING_SYSTEM;
    }

    psignal(p, sig);

    return RM_OK;
}

RM_STATUS os_get_current_time(
    U032 *sec,
    U032 *usec
)
{
    struct timeval tv;

    getmicrotime(&tv);

    *sec  = tv.tv_sec;
    *usec = tv.tv_usec;

    return RM_OK;
}

BOOL os_is_administrator(PHWINFO pDev)
{
    return suser(CURTHREAD);
}

U008 os_io_read_byte(
    PHWINFO pDev,
    U032 address
)
{
    /* XXX Fix me? (bus_space access) */
    return inb(address);
}

void os_io_write_byte(
    PHWINFO pDev,
    U032 address,
    U008 value
)
{
    /* XXX Fix me? (bus_space access) */
    outb(address, value);
}

U016 os_io_read_word(
    PHWINFO pDev,
    U032 address
)
{
    /* XXX Fix me? (bus_space access) */
    return inw(address);
}

void os_io_write_word(
    PHWINFO pDev,
    U032 address,
    U016 value
)
{
    /* XXX Fix me? (bus_space access) */
    return outw(address, value);
}

U032 os_io_read_dword(
    PHWINFO pDev,
    U032 address
)
{
    /* XXX Fix me? (bus_space access) */
    return inl(address);
}

void os_io_write_dword(
    PHWINFO pDev,
    U032 address,
    U032 value
)
{
    /* XXX Fix me? (bus_space access) */
    outl(address, value);
}

void* os_map_io_space(
    U032 start,
    U032 size,
    void **private,
    U032 user,
    U032 mode
)
{
    os_dbg_breakpoint();
    return NULL;
}

void os_unmap_io_space(
    void *address,
    U032 size,
    void *private,
    U032 user
)
{
    /* ignore calls from RM */
}

/*
 * This routine maps a range of physical pages into kernel virtual memory
 * to make it accessible by the kernel module. Normally, we wouldn't need
 * this, the kernel provides drivers with linear mappings for each of the
 * memory resources of the device the driver registered itself for.
 *
 * One of the aspects of this driver that makes this routine necessary is
 * the built-in AGP GART driver.
 */

void* os_map_kernel_space(
    U032 start,
    U032 size,
    U032 mode
)
{
    void *vm;
    vm_offset_t va, vt;
    pt_entry_t *pte;

    vm = pmap_mapdev(start, size);

    if (vm != NULL && mode == NV_MEMORY_UNCACHED) {
        va = vt = (vm_offset_t) vm;
        do {
            pte = vtopte(vt);
            pte_store(pte, pte_load(pte) | PG_N);
            vt += PAGE_SIZE;
        } while (vt < va + size);

        pmap_invalidate_range(kernel_pmap, va, vt);
    }

    /* XXX Fix me! (x86 only) */
    return vm;
}

void os_unmap_kernel_space(
    void *address,
    U032 size
)
{
    /* delete kmap and pmap */
    pmap_unmapdev((vm_offset_t) address, size);
}

RM_STATUS os_set_mem_range(
    U032 start,
    U032 size,
    U032 mode
)
{
    int arg;
    struct mem_range_desc mrd;

    mrd.mr_base  = start;
    mrd.mr_len   = size;
    mrd.mr_flags = MDF_WRITECOMBINE;

    strcpy(mrd.mr_owner, "NVIDIA");
    arg = MEMRANGE_SET_UPDATE;

    if (mem_range_attr_set(&mrd, &arg))
        return RM_ERROR;

    return RM_OK;
}

RM_STATUS os_unset_mem_range(
    U032 start,
    U032 size
)
{
    int arg;
    struct mem_range_desc mrd;

    mrd.mr_base = start;
    mrd.mr_len  = size;
    arg = MEMRANGE_SET_REMOVE;

    if (mem_range_attr_set(&mrd, &arg))
        return RM_ERROR;

    return RM_OK;
}


ULONG os_sti(ULONG flags)
{
    splx(flags);
    return flags; /* XXX Fix me? (sti, smp) */
}

ULONG os_cli(ULONG flags)
{
    flags = splhigh();
    return flags; /* XXX Fix me? (cli, smp) */
}

/*
 * The current debug level is used to determine if certain debug messages
 * are printed to the system console/log files or not. It defaults to the
 * highest debug level, i.e. the lowest debug output.
 */

int cur_debuglevel = 0xffffffff;

void os_dbg_init(void)
{
    U032 new_debuglevel;

    if (rm_read_registry_dword(NULL, "", "ResmanDebugLevel",
                &new_debuglevel) == RM_OK) {
        if (new_debuglevel != ~0)
            cur_debuglevel = new_debuglevel;
    }
}

void os_dbg_set_level(U032 new_debuglevel)
{
    cur_debuglevel = new_debuglevel;
}

void os_dbg_breakpoint(void)
{
#ifdef DEBUG
    /*
     * Enter the FreeBSD DDB kernel debugger; this interface is valid for
     * all debugger frontends this kernel may support. Of interest for us
     * is the GDB stub, which allows source-level remote debugging.
     */
    Debugger("os_dbg_breakpoint()");
#endif
}

#ifdef DEBUG
void out_string(const char *message)
{
    printf("%s", message);
}
#endif

static char nv_error_string[MAX_ERROR_STRING];

int nv_printf(
    int debuglevel,
    const char *format,
    ...
)
{
    char *message = nv_error_string;
    va_list arglist;
    int chars_written = 0;

    if (debuglevel >= ((cur_debuglevel >> 4) & 3)) {
        va_start(arglist, format);
        chars_written = vsprintf(message, format, arglist);
        va_end(arglist);
        printf("%s", message);
    }

    return chars_written;
}

S032 os_mem_cmp(
    const U008 *buf0,
    const U008 *buf1,
    U032 length
)
{
    return memcmp(buf0, buf1, length);
}

U008* os_mem_copy(
    U008 *dst,
    const U008 *src,
    U032 length
)
{
    return memcpy(dst, src, length);
}

RM_STATUS os_memcpy_from_user(
    void *dst,
    const void *src,
    U032 length
)
{
    if (src < (void *) VM_MAXUSER_ADDRESS)
        return copyin(src, dst, length)  ? RM_ERR_BAD_ADDRESS : RM_OK;

    return os_mem_copy(dst, src, length) ? RM_ERR_BAD_ADDRESS : RM_OK;
}

RM_STATUS os_memcpy_to_user(
    void *dst,
    const void *src,
    U032 length
)
{
    if (dst < (void *) VM_MAXUSER_ADDRESS)
        return copyout(src, dst, length) ? RM_ERR_BAD_ADDRESS : RM_OK;

    return os_mem_copy(dst, src, length) ? RM_ERR_BAD_ADDRESS : RM_OK;
}

void* os_mem_set(
    void *b,
    U008 c,
    U032 length
)
{
    return memset(b, c, length);
}

S032 os_string_compare(
    const U008 *s1,
    const U008 *s2
)
{
    return strcmp(s1, s2);
}

U008* os_string_copy(
    U008 *dst,
    const U008 *src
)
{
    return strcpy(dst, src);
}

U032 os_string_length(const U008* s)
{
    return strlen(s);
}

RM_STATUS os_strncpy_from_user(
    U008 *dst,
    const U008 *src,
    U032 n
)
{
    int done;

    return copyinstr(src, dst, n, &done) ? RM_ERR_BAD_ADDRESS : RM_OK;
}

U032 os_get_page_size(void)
{
    return PAGE_SIZE;
}

ULONG os_get_page_mask(void)
{
    /*
     * On FreeBSD, PAGE_MASK means (PAGE_SIZE - 1); on Linux it means the
     * opposite, ~(PAGE_SIZE - 1); that is what this function is expected
     * to return.
     */
    return ~PAGE_MASK;
}

U032 os_get_cpu_count(void)
{
#if defined(SMP)
    /*
     * XXX We need opt_global.h for the SMP and APIC_IO #define(s), which
     * is in ../../compile/SYSTEM_NAME (see config(8)). How do we get it?
     */
    return mp_ncpus;
#else
    return 1; /* single processor system */
#endif
}

RM_STATUS os_flush_cpu_cache(void)
{
    /*
     * XXX This will do for now, but this may need to be extended to make
     * IPI calls (flushing all caches).
     */
    __asm__ __volatile__("wbinvd": : :"memory");
    return RM_OK;
}

RM_STATUS os_raise_smp_barrier(void)
{
    return RM_OK;
}

RM_STATUS os_clear_smp_barrier(void)
{
    return RM_OK;
}

struct os_mutex {
#if __FreeBSD_version >= 500000
    struct mtx mutex_mtx;
#endif
    struct lock mutex_lock;
    int refcnt;
};

RM_STATUS os_alloc_sema(void **semaphore)
{
    struct os_mutex *mtx;
    RM_STATUS status;

    if ((status = os_alloc_mem((void **) &mtx, sizeof(struct os_mutex))
            != RM_OK))
        return status;

#if __FreeBSD_version >= 500000
    mtx_init(&mtx->mutex_mtx, "rm.mutex_mtx", NULL, MTX_SPIN | MTX_RECURSE);
#endif

    /*
     * Assume that none of the API paths are reentrant and
     * that any could could conceivably sleep.
     */
    lockinit(&mtx->mutex_lock, PZERO | PCATCH, "rm.mutex_lock", 0, 0);

    mtx->refcnt = 1;
    *semaphore = (void *) mtx;

    return RM_OK;
}

RM_STATUS os_free_sema(void *semaphore)
{
    struct os_mutex *mtx = semaphore;

#if __FreeBSD_version >= 500000
    mtx_destroy(&mtx->mutex_mtx);
#endif
    lockmgr(&mtx->mutex_lock, LK_DRAIN, 0, CURTHREAD);
    os_free_mem(semaphore);

    return RM_OK;
}

RM_STATUS os_acquire_sema(void *semaphore)
{
    struct os_mutex *mtx = semaphore;

#if __FreeBSD_version >= 500000
    mtx_lock_spin(&mtx->mutex_mtx);
    if (mtx->refcnt > 0)
        rm_disable_interrupts();
    mtx->refcnt--;
    mtx_unlock_spin(&mtx->mutex_mtx);
    lockmgr(&mtx->mutex_lock, LK_EXCLUSIVE, 0, CURTHREAD);
#else
    int spl = splhigh();
    if (mtx->refcnt > 0)
        rm_disable_interrupts();
    mtx->refcnt--;
    splx(spl);
    lockmgr(&mtx->mutex_lock, LK_EXCLUSIVE, 0, CURTHREAD);
#endif

    return RM_OK;
}

BOOL os_cond_acquire_sema(void *semaphore)
{
    struct os_mutex *mtx = semaphore;

#if __FreeBSD_version >= 500000
    mtx_lock_spin(&mtx->mutex_mtx);
    if (mtx->refcnt < 1) {
        mtx_unlock_spin(&mtx->mutex_mtx);
        return FALSE;
    } else {
        rm_disable_interrupts();
        mtx->refcnt--;
        mtx_unlock_spin(&mtx->mutex_mtx);
        lockmgr(&mtx->mutex_lock, LK_EXCLUSIVE, 0, CURTHREAD);
    }
#else
    int spl = splhigh();
    if (mtx->refcnt < 1) {
        splx(spl);
        return FALSE;
    } else {
        rm_disable_interrupts();
        mtx->refcnt--;
        splx(spl);
        lockmgr(&mtx->mutex_lock, LK_EXCLUSIVE, 0, CURTHREAD);
    }
#endif

    return TRUE;
}

RM_STATUS os_release_sema(void *semaphore)
{
    struct os_mutex *mtx = semaphore;

#if __FreeBSD_version >= 500000
    mtx_lock_spin(&mtx->mutex_mtx);
    if (!mtx->refcnt)
        rm_enable_interrupts();
    mtx->refcnt++;
    lockmgr(&mtx->mutex_lock, LK_RELEASE, 0, CURTHREAD);
    mtx_unlock_spin(&mtx->mutex_mtx);
#else
    int spl = splhigh();
    if (!mtx->refcnt)
        rm_enable_interrupts();
    mtx->refcnt++;
    lockmgr(&mtx->mutex_lock, LK_RELEASE, 0, CURTHREAD);
    splx(spl);
#endif

    return RM_OK;
}

BOOL os_is_acquired_sema(void *semaphore)
{
    struct os_mutex *mtx = semaphore;
    return (mtx->refcnt < 1);
}

#ifdef NV_CANARY
void os_canary_arm(void)
{
    /* XXX Implement me? (stack overflow check) */
}

void os_canary_check(void)
{
    /* XXX Implement me? (stack overflow check) */
}
#endif

